package util

import (
	"bytes"
	"io"
	"regexp"
)

type FilteredWriter struct {
	buffer  *bytes.Buffer
	writer  io.Writer
	pattern *regexp.Regexp
}

func (w FilteredWriter) processChunk(chunk []byte) error {
	index := 0
	for index < len(chunk) {
		n, err := w.writer.Write(chunk[index:])
		if err != nil {
			return err
		}
		index += n
	}
	return nil
}

func (w FilteredWriter) Flush() error {
	i := 0
	for {
		i++
		line, err := w.buffer.ReadBytes('\n')
		if err == io.EOF {
			return nil
		}
		if err != nil {
			return err
		}
		if !w.pattern.Match(line) {
			index := 0
			for index < len(line) {
				n, err := w.writer.Write(line[index:])
				index += n
				if err != nil {
					return err
				}
			}
		}
	}
}

func (w FilteredWriter) Close() error {
	if err := w.Flush(); err != nil {
		return err
	}
	if writeCloser, ok := w.writer.(io.WriteCloser); ok {
		return writeCloser.Close()
	}
	return nil
}

func (w FilteredWriter) Write(p []byte) (n int, err error) {
	if w.buffer.Available() < len(p) {
		err = w.Flush()
		if err != nil {
			return
		}
	}
	n, err = w.buffer.Write(p)
	return
}

func NewFilteredWriter(writer io.Writer, pattern *regexp.Regexp) FilteredWriter {
	return FilteredWriter{
		buffer:  bytes.NewBuffer(make([]byte, 65536)),
		writer:  writer,
		pattern: pattern,
	}
}
